module ULC where

import qualified Prelude

__ :: any
__ = Prelude.error "Logical or arity value used"

data Bool =
   True
 | False

data Nat =
   O
 | S Nat

add :: Nat -> Nat -> Nat
add n m =
  case n of {
   O -> m;
   S p -> S (add p m)}

sub :: Nat -> Nat -> Nat
sub n m =
  case n of {
   O -> n;
   S k -> case m of {
           O -> n;
           S l -> sub k l}}

data Star x r =
   Star_refl x
 | Star_R x x x (Star x r) r

star_rect :: (a1 -> a3) -> (a1 -> a1 -> a1 -> (Star a1 a2) -> a3 -> a2 -> a3)
             -> a1 -> a1 -> (Star a1 a2) -> a3
star_rect f f0 _ _ s =
  case s of {
   Star_refl x -> f x;
   Star_R x y z s0 r -> f0 x y z s0 (star_rect f f0 x y s0) r}

star_rec :: (a1 -> a3) -> (a1 -> a1 -> a1 -> (Star a1 a2) -> a3 -> a2 -> a3)
            -> a1 -> a1 -> (Star a1 a2) -> a3
star_rec =
  star_rect

eqb :: Nat -> Nat -> Bool
eqb n m =
  case n of {
   O -> case m of {
         O -> True;
         S _ -> False};
   S n' -> case m of {
            O -> False;
            S m' -> eqb n' m'}}

data Term =
   Var Nat
 | Abs Term
 | App Term Term

bltb :: Nat -> Nat -> Bool
bltb a b =
  case a of {
   O -> case b of {
         O -> False;
         S _ -> True};
   S x -> case b of {
           O -> False;
           S y -> bltb x y}}

rev_extensional_inference :: ()
rev_extensional_inference =
  __

decidable_nats :: ()
decidable_nats =
  __

decidable_terms :: ()
decidable_terms =
  __

data Term0 =
   Var0 Nat
 | Abs0 Term0
 | App0 Term0 Term0

data Star0 x r =
   Star_refl0 x
 | Star_R0 x x x (Star0 x r) r

substn :: Term0 -> Term0 -> Nat -> Term0
substn t s k =
  case t of {
   Var0 i ->
    case bltb k i of {
     True -> Var0 (sub i (S O));
     False -> case eqb i k of {
               True -> s;
               False -> Var0 i}};
   Abs0 u -> Abs0 (substn u s (add k (S O)));
   App0 u v -> App0 (substn u s k) (substn v s k)}

subst :: Term0 -> Term0 -> Term0
subst t s =
  substn t s (S O)

data Red =
   Bred Term0 Term
 | Appl Term0 Term0 Term Red
 | Appr Term0 Term0 Term Red
 | Abs1 Term0 Term0 Red

type Reds = Star0 Term0 Red

substn_eq :: ()
substn_eq =
  __

subst_gt :: ()
subst_gt =
  __

appl_reds :: Term0 -> Term0 -> Term -> Reds -> Reds
appl_reds m n x h =
  star_rec (\x0 -> Star_refl0 (App x0 x)) (\x0 y z _ iHstar r -> Star_R0 (App
    x0 x) (App y x) (App z x) iHstar (Appl y z x r)) m n h

appr_reds :: Term0 -> Term0 -> Term -> Reds -> Reds
appr_reds m n x h =
  star_rec (\x0 -> Star_refl0 (App x x0)) (\x0 y z _ iHstar r -> Star_R0 (App
    x x0) (App x y) (App x z) iHstar (Appr y z x r)) m n h

abs_reds :: Term0 -> Term0 -> Reds -> Reds
abs_reds m n h =
  star_rec (\x -> Star_refl0 (Abs0 x)) (\x y z _ iHstar r -> Star_R0 (Abs0 x)
    (Abs0 y) (Abs0 z) iHstar (Abs1 y z r)) m n h

breds :: Term0 -> Term0 -> Term0 -> Term0 -> Reds -> Reds -> Reds
breds m m' _ n' h _ =
  star_rec (\x -> Star_R0 (App (Abs0 x) n') (App (Abs0 x) n') (subst x n')
    (Star_refl0 (App (Abs0 x) n')) (Bred x n')) (\_ _ z _ _ _ -> Star_R0 (App
    (Abs0 z) n') (App (Abs0 z) n') (subst z n') (Star_refl0 (App (Abs0 z)
    n')) (Bred z n')) m m' h
