Module GRels.

Definition wf (A : Type) (R : A -> A -> Prop) :=
  well_founded (fun x y => R x y).

Definition dec_binary (A : Type) (R : A -> A -> Prop) :=
  forall x y, { R x y } + { ~ R x y }.

Section Star.

Variable X : Set.
Variable R : X -> X -> Prop.

Inductive star : X -> X -> Prop :=
  | star_refl : forall x, star x x
  | star_R    : forall x y z, star x y -> R y z -> star x z.

Lemma R_to_star : forall x y, R x y -> star x y.
Proof.
  intros.
  apply (star_R _ x _).
  apply star_refl.
  exact H.
Qed.

Lemma star_trans : forall x y z, star x y -> star y z -> star x z.
Proof.
  intros.
  induction H0.
  exact H.
  apply (star_R _ y _).
  apply IHstar.
  exact H.
  exact H1.
Qed.

End Star.

End GRels.