Require Import FunctionalExtensionality.
Require Import SK.
Require Import SK_currycombinators.
Require Import SK_wreduction.
Require Import UntypedLC.
Require Import UntypedLC_aux.

Include CL.

Fixpoint cl_to_lc (t : Term) : LC.Term :=
  match t with
  | S => LC.Abs (LC.Abs (LC.Abs (LC.App
    (LC.App (LC.Var 3) (LC.Var 1))((LC.App (LC.Var 2) (LC.Var 1))))
  ))
  | K => LC.Abs (LC.Abs (LC.Var 2))
  | M @ N => LC.App (cl_to_lc M) (cl_to_lc N)
  | ` s => lterm_to_term (LC.LVar s)
  end.

Fixpoint lc_to_cl (t : LC.Term) : Term :=
  match t with
  | LC.Abs (LC.Var 1) => I
  | LC.Abs (LC.App M N) => S @ (lc_to_cl M) @ (lc_to_cl N)
  | LC.Var x => ` "x"
  | LC.App M N => (lc_to_cl M) @ (lc_to_cl N)
  | LC.Abs M => K @ (lc_to_cl M)
  end.

Lemma skkx_equals_sksx : forall x, reds (S @ K @ K @ x) = reds (S @ K @ S @ x).
Proof.
  intros.
  rewrite s_schema. rewrite s_schema.
  rewrite k_schema. rewrite k_schema.
  reflexivity.
Qed.

Lemma skkx_eq_sksx : forall x, (S @ K @ K) @ x = (S @ K @ S) @ x.
Proof.
  intros.
  rewrite s_schema.
  rewrite s_schema.
  rewrite k_schema.
  rewrite k_schema.
  reflexivity.
Qed.

Lemma skk_eq_sks : (forall x, 
  S @ K @ K @ x = S @ K @ S @ x) -> App (S @ K @ K) = App (S @ K @ S).
Proof.
  intros.
  apply functional_extensionality in H.
  exact H.
Qed.