Require Import General_relations.
Require Import SK.

Include GRels.
Include CL.

Set Implicit Arguments.

Axiom s_schema : forall M N P, S @ M @ N @ P = M @ P @ (N @ P).
Axiom k_schema : forall M N, K @ M @ N = M.

Inductive red : Term -> Term -> Prop :=
  | s_red : forall M N P, red (S @ M @ N @ P) (M @ P @ (N @ P))
  | k_red : forall M N, red (K @ M @ N) M
  | appl_red : forall M N X, red M N -> red (M @ X) (N @ X)
  | appr_red : forall M N X, red M N -> red (X @ M) (X @ N)
  | refl : forall M, red M M.

Definition reds := star Term red.
Definition redt := star_trans Term red.

Lemma appl_reds : forall M N X, reds M N -> reds (M @ X) (N @ X).
Proof.
  intros.
  induction H.
  apply star_refl.
  apply (star_R _ _ _ (y @ X) _).
  exact IHstar.
  apply appl_red.
  exact H0.
Qed.

Lemma appr_reds : forall M N X, reds M N -> reds (X @ M) (X @ N).
Proof.
  intros.
  induction H.
  apply star_refl.
  apply (star_R _ _ _ (X @ y) _).
  exact IHstar.
  apply appr_red.
  exact H0.
Qed.

Lemma app_reds : forall M M' N N', reds M N -> reds M' N' -> reds (M @ M') (N @ N').
Proof.
  intros.
  apply (redt (appl_reds _ H) (appr_reds _ H0)).
Qed.

Lemma S_reds : forall M M' N N' P P', reds M M' -> reds N N' -> reds P P'
  -> reds (S @ M @ N @ P) (M' @ P' @ (N' @ P')).
Proof.
  intros.
  apply (star_R _ _ _ (S @ M' @ N' @ P') _).
  apply app_reds.
  apply app_reds.
  apply appr_reds.
  exact H.
  exact H0.
  exact H1.
  apply s_red.
Qed.

Lemma K_reds : forall M M' N, reds M M' -> reds (K @ M @ N) M'.
Proof.
  intros. apply (star_R _ _ _ (K @ M' @ N) _).
  apply appl_reds.
  apply appr_reds.
  exact H.
  apply k_red.
Qed.