Require Import Nat.
Require Import String.

Module LC.

Set Implicit Arguments.

(* 
  Términos del cálculo lambda no tipado
  utilizando índices de de Bruijn.
*)

Inductive Term : Set :=
  | Var : nat -> Term
  | Abs : Term -> Term
  | App : Term -> Term -> Term.

Infix "@" := App (left associativity, at level 51).

(*
  Términos del cálculo lambda no tipado
  utilizando la notación común.

  Nótese que las demostraciones y definiciones
  relevantes al rededor del cálculo lambda
  no tipado se hacen para los términos dados
  con índices de de Bruijn.
*)

Inductive LTerm : Type :=
  | LVar : string -> LTerm
  | LAbs : string -> LTerm -> LTerm
  | LApp : LTerm -> LTerm -> LTerm.

(*
  Notación auxiliar para lambda términos comunes.
*)

Notation "` x" := (LVar x) (at level 20).
Notation "'λ' x M" := (LAbs x M) (at level 30, right associativity).
Infix "#" := LApp (at level 25, left associativity).

(*
  Función que computa la igualdad entre dos términos
  del cálculo lambda no tipado con índices de de Bruijn.
*)

Fixpoint eqterm n m :=
  match n, m with
  | Var n, Var m => eqb n m
  | Abs n, Abs m => eqterm n m
  | App nl nr, App ml mr => andb (eqterm nl ml) (eqterm nr mr)
  | _, _ => false
  end.


(*
  Tamaño del árbol que representa a una lambda expresión.
*)

Fixpoint size t :=
  match t with
  | Var i => 1
  | App u v => (size u) + (size v)
  | Abs u => 1 + (size u)
  end.

Definition lnat_0 := Abs (Abs (Var 1)).

Definition lnat_succ (n : Term) : Term :=
  Abs (Abs (Abs (Var 2) @((Var 3) @ (Var 1)))).

Definition lnat_sum (n m : Term) : Term :=
  Abs (Abs (((m @ (Var 2)) @ (n @ (Var 2))) @ (Var 1))).

Definition lnat_mult (n m : Term) : Term :=
  Abs (m @ (n @ (Var 1))).

End LC.